#include "VulkanUtils.hpp"

#include <core/Array.hpp>
#include <core/File.hpp>
#include <core/Logger.hpp>

#include "core/VulkanUtility.hpp"

static u32 getSwapImageCount(const VkSurfaceCapabilitiesKHR* caps) {
    u32 c = caps->minImageCount + 1;
    // according to VkSurfaceCapabilitiesKHR doc:
    // maxImageCount is 0 when there is no strict limit
    if(caps->maxImageCount != 0) {
        return Core::min(c, caps->maxImageCount);
    }
    return c;
}

bool initVulkanSwapchain(VkSwapchainKHR* sc, VulkanSwapchainData* d) {
    VkSurfaceCapabilitiesKHR caps = {0};
    if(d->base.getSurfaceCapabilities(caps)) {
        return true;
    }
    VkSwapchainCreateInfoKHR ci = {
        .sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
        .surface = d->base,
        .minImageCount = getSwapImageCount(&caps),
        .imageFormat = d->surfaceFormat.format,
        .imageColorSpace = d->surfaceFormat.colorSpace,
        .imageExtent = d->size,
        .imageArrayLayers = 1,
        .imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
        .imageSharingMode = d->sharingMode,
        .queueFamilyIndexCount = static_cast<u32>(d->queueFamilies.getLength()),
        .pQueueFamilyIndices = &d->queueFamilies[0],
        .preTransform = caps.currentTransform,
        .compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR,
        .presentMode = d->presentMode,
        .clipped = VK_TRUE,
        .oldSwapchain = VK_NULL_HANDLE};
    VK_CHECK_TRUE(vkCreateSwapchainKHR(d->base, &ci, nullptr, sc));
    return false;
}

void destroyVulkanSwapchain(VkSwapchainKHR s, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroySwapchainKHR(d, s, nullptr);
    }
}

static bool createImageView(
    VkImageView& view, VkDevice d, VkImage image, VkFormat format) {
    VkImageViewCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
        .image = image,
        .viewType = VK_IMAGE_VIEW_TYPE_2D,
        .format = format,
        .subresourceRange = {
            .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
            .baseMipLevel = 0,
            .levelCount = 1,
            .baseArrayLayer = 0,
            .layerCount = 1}};
    VK_CHECK_TRUE(vkCreateImageView(d, &info, nullptr, &view));
    return false;
}

bool initVulkanSwapchainImages(
    VulkanSwapchainImages& si, VkDevice d, VkSwapchainKHR sc, VkFormat format) {
    u32 c = 0;
    VK_CHECK_TRUE(vkGetSwapchainImagesKHR(d, sc, &c, nullptr));
    si.images.resize(c);
    si.imageViews.resize(c);
    VK_CHECK_TRUE(vkGetSwapchainImagesKHR(d, sc, &c, &si.images[0]));
    for(u32 x = 0; x < c; x++) {
        if(createImageView(si.imageViews[x], d, si.images[x], format)) {
            return true;
        }
    }
    return false;
}

void destroyVulkanSwapchainImages(VulkanSwapchainImages& si, VkDevice d) {
    if(d == VK_NULL_HANDLE) {
        return;
    }
    for(VkImageView& v : si.imageViews) {
        vkDestroyImageView(d, v, nullptr);
    }
    si.imageViews.clear();
    si.images.clear();
}

static bool initShaderModule(
    Core::List<char>& f, VkDevice d, VkShaderModule* sm) {
    size_t l = f.getLength() - 1;
    if((l % 4) != 0) {
        VK_REPORT_ERROR("Shader size is not a multiple of 4");
        return true;
    }
    VkShaderModuleCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
        .codeSize = l,
        .pCode = reinterpret_cast<u32*>(static_cast<void*>(&f[0]))};
    VK_CHECK_TRUE(vkCreateShaderModule(d, &info, nullptr, sm));
    return false;
}

bool initVulkanShaderModule(VkShaderModule* sm, VkDevice d, const char* path) {
    Core::List<char> f;
    if(Core::readFile(f, path)) {
        return true;
    }
    return initShaderModule(f, d, sm);
}

void destroyVulkanShaderModule(VkShaderModule sm, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroyShaderModule(d, sm, nullptr);
    }
}

bool initVulkanPipelineLayout(VkPipelineLayout* pl, VkDevice d) {
    VkPipelineLayoutCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO,
        .setLayoutCount = 0,
        .pSetLayouts = nullptr,
        .pushConstantRangeCount = 0,
        .pPushConstantRanges = nullptr};
    VK_CHECK_TRUE(vkCreatePipelineLayout(d, &info, nullptr, pl));
    return false;
}

void destroyVulkanPipelineLayout(VkPipelineLayout pl, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroyPipelineLayout(d, pl, nullptr);
    }
}

bool initVulkanFramebuffers(
    Core::List<VkFramebuffer>& f, VulkanSwapchainImages& si, VkDevice d,
    VkRenderPass rp, u32 width, u32 height) {
    f.resize(si.imageViews.getLength());
    for(u32 i = 0; i < si.imageViews.getLength(); i++) {
        VkFramebufferCreateInfo info = {
            .sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
            .renderPass = rp,
            .attachmentCount = 1,
            .pAttachments = &si.imageViews[i],
            .width = width,
            .height = height,
            .layers = 1};
        VK_CHECK_TRUE(vkCreateFramebuffer(d, &info, nullptr, &f[i]));
    }
    return false;
}

void destroyVulkanFramebuffers(Core::List<VkFramebuffer>& fs, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        for(VkFramebuffer f : fs) {
            vkDestroyFramebuffer(d, f, nullptr);
        }
    }
    fs.clear();
}

bool initCommandVulkanBuffer(
    VkCommandBuffer* cb, VkDevice d, VkCommandPool cp) {
    VkCommandBufferAllocateInfo info = {
        .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
        .commandPool = cp,
        .level = VK_COMMAND_BUFFER_LEVEL_PRIMARY,
        .commandBufferCount = 1};
    VK_CHECK_TRUE(vkAllocateCommandBuffers(d, &info, cb));
    return false;
}

bool initVulkanSemaphore(VkSemaphore* s, VkDevice d) {
    VkSemaphoreCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO};
    VK_CHECK_TRUE(vkCreateSemaphore(d, &info, nullptr, s));
    return false;
}

void destroyVulkanSemaphore(VkSemaphore s, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroySemaphore(d, s, nullptr);
    }
}

bool initVulkanFence(VkFence* f, VkDevice d) {
    VkFenceCreateInfo info = {
        .sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO,
        .flags = VK_FENCE_CREATE_SIGNALED_BIT};
    VK_CHECK_TRUE(vkCreateFence(d, &info, nullptr, f));
    return false;
}

void destroyVulkanFence(VkFence f, VkDevice d) {
    if(d != VK_NULL_HANDLE) {
        vkDestroyFence(d, f, nullptr);
    }
}
